import os
import unittest
import logging
import shutil
import string
import gc
import json
from copy import copy
import random
import time
from pathlib import Path
from unittest.mock import MagicMock
import warnings

# prevent excessive warning logs
warnings.filterwarnings('ignore')
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import pytest

import tensorflow as tf
import pandas as pd
import numpy as np
from sklearn.metrics import accuracy_score, recall_score
from sklearn.model_selection import train_test_split

from finetune import Classifier, Regressor, DeploymentModel, SequenceLabeler, ComparisonRegressor
from finetune.base_models import GPTModel
from finetune.datasets import generic_download
from finetune.config import get_config
from finetune.errors import FinetuneError

SST_FILENAME = "SST-binary.csv"

class TestDeploymentModel(unittest.TestCase):
    n_sample = 20
    do_comparison = True
    base_model = GPTModel

    classifier_path = "tests/saved-models/deployment_classifier.jl"
    comparison_regressor_path = "tests/saved-models/deployment_comparison_regressor.jl"
    sequence_labeler_path = "tests/saved-models/deployment_sequence_labeler.jl"
    classifier_dataset_path = os.path.join(
        'Data', 'Classify', 'SST-binary.csv'
    )
    sequence_dataset_path = os.path.join(
        'Data', 'Sequence', 'reuters.xml'
    )
    processed_path = os.path.join('Data', 'Sequence', 'reuters.json')

    @classmethod
    def _download_data(cls):
        
        #Download Stanford Sentiment Treebank to data directory
        
        path = Path(cls.classifier_dataset_path)
        if path.exists():
            return

        path.parent.mkdir(parents=True, exist_ok=True)
        generic_download(
            url="https://s3.amazonaws.com/enso-data/SST-binary.csv",
            text_column="Text",
            target_column="Target",
            filename=SST_FILENAME
        )

        #Download Reuters Dataset to enso `data` directory
        
        path = Path(cls.sequence_dataset_path)
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)

        if not os.path.exists(cls.sequence_dataset_path):
            url = "https://raw.githubusercontent.com/dice-group/n3-collection/master/reuters.xml"
            r = requests.get(url)
            with open(cls.sequence_dataset_path, "wb") as fp:
                fp.write(r.content)

        with codecs.open(cls.sequence_dataset_path, "r", "utf-8") as infile:
            soup = bs(infile, "html.parser")

        docs = []
        docs_labels = []
        for elem in soup.find_all("document"):
            texts = []
            labels = []

            # Loop through each child of the element under "textwithnamedentities"
            for c in elem.find("textwithnamedentities").children:
                if type(c) == Tag:
                    if c.name == "namedentityintext":
                        label = "Named Entity"  # part of a named entity
                    else:
                        label = "<PAD>"  # irrelevant word
                    texts.append(c.text)
                    labels.append(label)

            docs.append(texts)
            docs_labels.append(labels)


        with open(cls.processed_path, 'wt') as fp:
            json.dump((docs, docs_labels), fp)

    @classmethod
    def setUpClass(cls):
        cls._download_data()
        
        #dataset preparation
        cls.classifier_dataset = pd.read_csv(cls.classifier_dataset_path, nrows=cls.n_sample * 10)

        path = os.path.join(os.path.dirname(__file__), "data", "testdata.json")
        with open(path, 'rt') as fp:
            cls.texts, cls.labels = json.load(fp)

        cls.animals = ["dog", "cat", "horse", "cow", "pig", "sheep", "goat", "chicken", "guinea pig", "donkey", "turkey", "duck", "camel", "goose", "llama", "rabbit", "fox"]
        cls.numbers = ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen", "sixteen"]
        
        #train and save sequence labeler for later use
        try:
            cls.s = SequenceLabeler.load(cls.sequence_labeler_path, **cls.default_seq_config(cls))
        except FileNotFoundError:
            cls.s = SequenceLabeler(**cls.default_seq_config(cls))
            cls.s.fit(cls.texts * 10, cls.labels * 10)
            cls.s.save(cls.sequence_labeler_path)
        
        #train and save classifier for later use
        train_sample = cls.classifier_dataset.sample(n=cls.n_sample*10)
        try:
            cls.cl = Classifier.load(cls.classifier_path)
        except FileNotFoundError:
            cls.cl = Classifier(**cls.default_config(cls))
            cls.cl.fit(train_sample.Text, train_sample.Target)
            cls.cl.save(cls.classifier_path)

        if cls.do_comparison:
            #train and save comparison regressor for use
            cls.cr = ComparisonRegressor()
    
            n_per = 150
            similar = []
            different = []
            for dataset in [cls.animals, cls.numbers]:
                for i in range(n_per // 2):
                    similar.append([random.choice(dataset), random.choice(dataset)])
            for i in range(n_per):
                different.append([random.choice(cls.animals), random.choice(cls.numbers)])

            targets = np.asarray([1] * len(similar) + [0] * len(different))
            data = similar + different

            cls.x_tr, cls.x_te, cls.t_tr, cls.t_te = train_test_split(data, targets, test_size=0.3, random_state=42)
            
            try:
                cls.cr = ComparisonRegressor.load(cls.comparison_regressor_path, **cls.default_config(cls))
            except FileNotFoundError:
                cls.cr = ComparisonRegressor(**cls.default_config(cls))
                cls.cr.fit(cls.x_tr, cls.t_tr)
                cls.cr.save(cls.comparison_regressor_path)

    def setUp(self):
        try:
            os.mkdir("tests/saved-models")
        except FileExistsError:
            warnings.warn("tests/saved-models still exists, it is possible that some test is not cleaning up properly.")

    @classmethod  
    def tearDownClass(cls):
        shutil.rmtree("tests/saved-models/")

    def default_config(cls, **kwargs):
        defaults = {
            'batch_size': 2,
            'max_length': 256,
            'n_epochs': 3,
            'adapter_size': 64,
            'chunk_long_sequences': False
        }
        defaults.update(kwargs)
        return defaults

    def default_seq_config(cls, **kwargs):
        d = dict(
            batch_size=2,
            max_length=256,
            lm_loss_coef=0.0,
            val_size=0,
            adapter_size=64,
            base_model=cls.base_model,
            interpolate_pos_embed=False,
        )
        d.update(**kwargs)
        return d

    def default_comp_config(cls, **kwargs):
        d = dict(
            batch_size=2,
            max_length=256,
            n_epochs=1,
            l2_reg=0,
            lm_loss_coef=0.,
            val_size=0.,
        )
        d.update(kwargs)
        return d

    def test_cached_predict(self):
        """
        Ensure second call to predict is faster than first, including loading.
        """
        model = DeploymentModel(featurizer=self.base_model, **self.default_config())
        model.load_featurizer()

        valid_sample = self.classifier_dataset.sample(n=self.n_sample)
        start = time.time()
        model.load_custom_model(self.classifier_path)
        model.predict(valid_sample.Text[:1].values)
        first = time.time()
        model.predict(valid_sample.Text[:1].values)
        second = time.time()

        first_prediction_time = (first - start)
        second_prediction_time = (second - first)
        self.assertLess(second_prediction_time, first_prediction_time)
        model.close()

    def test_switching_models(self):
        """
        Ensure model can switch out weights without erroring out
        """
        model = DeploymentModel(featurizer=self.base_model, **self.default_config())
        model.load_featurizer()
        #test transitioning from any of [sequence labeling, comparison, default] to any other
        model.load_custom_model(self.classifier_path)
        if self.do_comparison:
            model.load_custom_model(self.comparison_regressor_path)
        model.load_custom_model(self.sequence_labeler_path)
        model.load_custom_model(self.classifier_path)
        model.load_custom_model(self.sequence_labeler_path)
        if self.do_comparison:
            model.load_custom_model(self.comparison_regressor_path)
        model.load_custom_model(self.classifier_path)
        model.close()

    def test_reasonable_predictions(self):
        """
        Ensure model produces reasonable predictions after loading weights
        """
        model = DeploymentModel(featurizer=self.base_model, **self.default_seq_config())
        model.load_featurizer()
        
        #test same output as weights loaded with Classifier model
        valid_sample = self.classifier_dataset.sample(n=self.n_sample)
        model.load_custom_model(self.classifier_path)
        deployment_preds = model.predict_proba(valid_sample.Text.values)
        model.close()
        classifier_preds = self.cl.predict_proba(valid_sample.Text.values)
        
        for c_pred, d_pred in zip(classifier_preds, deployment_preds):
            self.assertTrue(list(c_pred.keys()) == list(d_pred.keys()))
            for c_pred_val, d_pred_val in zip(c_pred.values(), d_pred.values()):
                np.testing.assert_almost_equal(c_pred_val, d_pred_val, decimal=4)
        
        if self.do_comparison:
            #test same output as weights loaded with Comparison Regressor model
            model = DeploymentModel(featurizer=self.base_model, **self.default_seq_config())
            model.load_featurizer()
            model.load_custom_model(self.comparison_regressor_path)
            deployment_preds = model.predict(self.x_te)
            model.close()
            compregressor = ComparisonRegressor.load(self.comparison_regressor_path,  **self.default_comp_config())
            compregressor_preds = compregressor.predict(self.x_te)
            for c_pred, d_pred in zip(compregressor_preds, deployment_preds):
                np.testing.assert_almost_equal(c_pred, d_pred, decimal=4)

    def test_large_predict(self):
        """
        Ensure model does not have OOM issues with large inputs for inference
        """
        large_dataset = self.animals*100
        model = DeploymentModel(featurizer=self.base_model, **self.default_config())
        model.load_featurizer()
        model.load_custom_model(self.classifier_path)
        model.predict(large_dataset)
        model.close()

    def test_fast_switch(self):
        """
        Ensure model can load/reload weights and predict in reasonable time
        """
        model = DeploymentModel(featurizer=self.base_model, **self.default_config())
        model.load_featurizer()
        model.load_custom_model(self.classifier_path)
        model.predict(['finetune'])

        start = time.time()
        model.load_custom_model(self.sequence_labeler_path)
        predictions = model.predict(['finetune sequence'])
        end = time.time()
        self.assertGreater(2.5, end - start)

        if self.do_comparison:
            start = time.time()
            model.load_custom_model(self.comparison_regressor_path)
            predictions = model.predict([['finetune', 'compare']])
            end = time.time()
            self.assertGreater(2.5, end - start)
        model.close()



import os
import unittest
import logging
import shutil
import random
import copy
import requests
import json
import pytest
import spacy
import warnings

# prevent excessive warning logs
warnings.filterwarnings("ignore")
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

import tensorflow as tf
import pandas as pd
import numpy as np

from sklearn.metrics import accuracy_score, recall_score
from sklearn.model_selection import train_test_split

from finetune import Classifier, SequenceLabeler
from finetune.base_models import TextCNN, BERTModelCased, GPT2Model, GPTModel, RoBERTa, GPT
from finetune.config import get_config
from finetune.util.metrics import (
    sequence_labeling_token_precision,
    sequence_labeling_token_recall,
)
from finetune.datasets.reuters import Reuters


class TestAuxiliary(unittest.TestCase):
    base_model = GPT
    default_context = {"token": "", "pos": "filler", "capitalized": "False"}

    @classmethod
    def create_context(cls, nlp, text, default):
        to_search = text
        removed = 0
        doc = nlp(text)
        context = []
        for token in doc:
            token_context = copy.deepcopy(default)
            start = to_search.find(token.text)
            assert start != -1
            to_search = to_search[start + len(token.text) :]
            end = start + len(token.text)
            token_context = {}
            token_context["token"] = token.text
            token_context["pos"] = token.pos_ if token.pos_ != "PROPN" else "NOUN"
            token_context["capitalized"] = (
                "False" if token.text.lower() == token.text else "True"
            )
            token_context["start"] = start + removed
            token_context["end"] = end + removed
            context.append(token_context)
            removed += end
        return context

    @classmethod
    def setUpClass(cls):
        nlp = spacy.load("en_core_web_sm")

        random.seed(42)
        np.random.seed(42)
        dataset = Reuters().dataframe
        dataset["annotations"] = [
            json.loads(annotation) for annotation in dataset["annotations"]
        ]
        trainX, testX, trainY, testY = train_test_split(
            dataset.texts.values,
            dataset.annotations.values,
            test_size=0.3,
            random_state=42,
        )

        train_context, test_context = [], []
        for text in trainX:
            context = cls.create_context(nlp, text, cls.default_context)
            train_context.append(context)
        trainX = [trainX, train_context]

        for text in testX:
            context = cls.create_context(nlp, text, cls.default_context)
            test_context.append(context)
        testX = [testX, test_context]

        cls.dataset = (trainX, testX, trainY, testY)

    def setUp(self):
        try:
            os.mkdir("tests/saved-models")
        except FileExistsError:
            warnings.warn(
                "tests/saved-models still exists, it is possible that some test is not cleaning up properly."
            )
            pass

    def tearDown(self):
        shutil.rmtree("tests/saved-models/")

    def default_config(self, **kwargs):
        defaults = {
            "batch_size": 2,
            "max_length": 256,
            "n_epochs": 3,
            "base_model": self.base_model,
            "val_size": 0,
            "default_context": self.default_context
        }
        defaults.update(kwargs)
        return dict(get_config(**defaults))
    
    def test_auxiliary_classifier(self):
        """
        Ensure model training does not error out
        Ensure model returns predictions
        """
        (trainX, testX, trainY, _) = self.dataset
        trainY = [
            random.randint(0, 1) for _ in range(len(trainY))
        ]  # random labels just to make sure there are no errors -> reasonable predictions tests are in sequence_label
        model = Classifier(
            **self.default_config()
        )
        model.fit(trainX, trainY)
        _ = model.predict(testX)
    
    def test_auxiliary_sequence_labeler(self):
        """
        Ensure model training does not error out
        Ensure model returns reasonable predictions
        """
        (trainX, testX, trainY, testY) = self.dataset
        model = SequenceLabeler(
            **self.default_config()
        )
        model.fit(trainX, trainY)
        preds = model.predict(testX)
        token_precision = sequence_labeling_token_precision(preds, testY)
        token_recall = sequence_labeling_token_recall(preds, testY)
        self.assertIn("Named Entity", token_precision)
        self.assertIn("Named Entity", token_recall)
        token_precision = np.mean(list(token_precision.values()))
        token_recall = np.mean(list(token_recall.values()))
        self.assertGreater(token_precision, 0.6)
        self.assertGreater(token_recall, 0.6)

    
    def test_save_load(self):
        """
        Ensure saving + loading does not cause errors
        Ensure saving + loading does not change predictions
        """
        save_file = "tests/saved-models/test-save-load"
        config = self.default_config(save_adam_vars=False, n_epochs=1)
        model = Classifier(**config)

        (trainX, testX, trainY, _) = self.dataset
        trainY = [random.randint(0, 1) for _ in range(len(trainY))]
        model.fit(trainX, trainY)
        predictions = model.predict(testX)
        model.save(save_file)

        model = Classifier.load(save_file)
        new_predictions = model.predict(testX)
        for i, prediction in enumerate(predictions):
            self.assertEqual(prediction, new_predictions[i])
    
class TestAuxiliaryBert(TestAuxiliary):
    base_model = BERTModelCased


class TestAuxiliaryGPT(TestAuxiliary):
    base_model = GPTModel


class TestAuxiliaryGPT2(TestAuxiliary):
    base_model = GPT2Model


class TestAuxiliaryRoberta(TestAuxiliary):
    base_model = RoBERTa
                                



import tensorflow as tf
from finetune.nn.nn_utils import dropout, norm


def add_auxiliary(context, context_dim, clf_h, seq_feats, config, train):
    context.set_shape([None, config.max_length, context_dim])
    context_embed_weights = tf.get_variable(
        name="ce",
        shape=[context_dim, config.n_context_embed],
        initializer=tf.random_normal_initializer(stddev=config.weight_stddev),
    )

    context_weighted_avg = tf.get_variable(
        name="cwa",
        shape=[context_dim],
        initializer=tf.random_normal_initializer(stddev=config.weight_stddev),
    )

    if config.train_embeddings:
        context_embed_weights = dropout(
            context_embed_weights, config.embed_p_drop, train
        )
        context_weighted_avg = dropout(context_weighted_avg, config.embed_p_drop, train)
    else:
        context_embed_weights = tf.stop_gradient(context_embed_weights)

    with tf.variable_scope("context_embedding"):
        weighted_C = tf.multiply(
            context, context_weighted_avg
        )  # [batch_size, seq_length, context_dim] * [context_dim] = [batch_size, seq_length, context_dim], with weighted inputs
        c_embed = tf.tensordot(
            weighted_C, context_embed_weights, axes=[[2], [0]]
        )  # [batch_size, seq_length, context_dim] * [context_dim, n_embed] = [batch_size, seq_length, n_embed]
        c_embed = norm(c_embed, tf.get_variable_scope())
        seq_feats = tf.concat([seq_feats, c_embed], axis=2)
        c_embed = tf.reduce_mean(c_embed, axis=1)
        clf_h = tf.concat([clf_h, c_embed], axis=1)
        return clf_h, seq_feats



import os
import unittest

# required for tensorflow logging control
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import pytest

from sklearn.model_selection import train_test_split
from finetune import Association

from finetune.config import get_config
from finetune.datasets.treebank_association import TreebankNounVP

SKIP_LM_TESTS = get_config().base_model.is_bidirectional


class TestAssociation(unittest.TestCase):

    @classmethod
    def get_data_and_schema(cls, rows=None):
        data = TreebankNounVP()
        return (
            data.get_data(rows),
            {
                "association_types": ["has_verb"],
                "viable_edges": {
                    "noun_phrase": [
                        ["verb", "has_verb"],
                    ]
                    , "verb": [None]
                }
            }
        )

    def setUp(self):
        data, schema = self.get_data_and_schema(20)
        self.texts, self.labels = data
        self.model = Association(batch_size=2, max_length=32, **schema)

    @pytest.mark.skipif(SKIP_LM_TESTS, reason="Bidirectional models do not yet support LM functions")
    def test_fit_lm_predict(self):
        """
        Ensure model training does not error out
        Ensure model returns predictions
        """
        train_texts, test_texts, train_annotations, test_annotations = train_test_split(self.texts, self.labels, test_size=0.1)
        self.model.fit(train_texts)
        self.model.fit(train_texts, train_annotations)
        predictions = self.model.predict(test_texts)
        self.assertIsInstance(predictions, list)
        self.assertEqual(len(predictions), len(test_texts))
        self.assertIsInstance(predictions[0], list)
        self.assertGreater(sum(len(p) for p in predictions), 0)
        # make sure there is at least one prediction, if there isn't somethings probably broken.

        for pred in predictions:
            pred_has_verb = any(p_i["label"] == "verb" for p_i in pred)
            for p_i in pred:
                self.assertIsInstance(p_i, dict)
                self.assertIsInstance(p_i["label"], str)
                self.assertIn(p_i["label"], ["verb", "noun_phrase"])

                if p_i["label"] == "noun_phrase" and pred_has_verb:
                    self.assertIn("associations", p_i)
                    # a noun_phrase MUST have a verb by the above schema.

                    self.assertEqual(p_i["associations"]["relationship"], "has_verb")
                    # has_verb is the only viable edge for a noun_phrase.

                    self.assertIn("prob", p_i["associations"])
                    self.assertTrue(0.0 < p_i["associations"]["prob"] < 1.0)
                    self.assertEqual(pred[p_i["associations"]["index"]]["label"], "verb")
                    # check it is associated with a verb
                else:
                    self.assertNotIn("associations", p_i)
                    # Verbs cannot have any associations by the above schema

            self.model.predict(test_texts)

    def test_fit_predict(self):
        """
        Ensure model training does not error out
        Ensure model returns predictions
        """
        train_texts, test_texts, train_annotations, test_annotations = train_test_split(
            self.texts, self.labels, test_size=0.1
        )
        self.model.fit(train_texts, train_annotations)
        predictions = self.model.predict(test_texts)
        self.assertIsInstance(predictions, list)
        self.assertEqual(len(predictions), len(test_texts))
        self.assertIsInstance(predictions[0], list)
        self.assertGreater(sum(len(p) for p in predictions), 0)
        # make sure there is at least one prediction, if there isn't somethings probably broken.

        for pred in predictions:
            pred_has_verb = any(p_i["label"] == "verb" for p_i in pred)
            for p_i in pred:
                self.assertIsInstance(p_i, dict)
                self.assertIsInstance(p_i["label"], str)
                self.assertIn(p_i["label"], ["verb", "noun_phrase"])

                if p_i["label"] == "noun_phrase" and pred_has_verb:
                    self.assertIn("associations", p_i)
                    # a noun_phrase MUST have a verb by the above schema.

                    self.assertEqual(p_i["associations"]["relationship"], "has_verb")
                    # has_verb is the only viable edge for a noun_phrase.

                    self.assertIn("prob", p_i["associations"])
                    self.assertTrue(0.0 < p_i["associations"]["prob"] < 1.0)
                    self.assertEqual(pred[p_i["associations"]["index"]]["label"], "verb")
                    # check it is associated with a verb
                else:
                    self.assertNotIn("associations", p_i)
                    # Verbs cannot have any associations by the above schema

            self.model.predict(test_texts)




import itertools
import logging
import sys
import math
import os
from collections import Counter

from abc import ABCMeta, abstractmethod

import tqdm
import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.python.data import Dataset
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelBinarizer
import finetune
from finetune.errors import FinetuneError
from finetune.encoding.input_encoder import ArrayEncodedOutput, EncodedOutput
from finetune.util.imbalance import compute_class_weights

LOGGER = logging.getLogger("finetune")


class BasePipeline(metaclass=ABCMeta):
    def __init__(self, config):
        self.config = config
        self.text_encoder = self.config.base_model.get_encoder()
        self.label_encoder = None
        self.target_dim = None
        self.pad_idx_ = None
        self.rebuild = False
        self.epoch = 0

    @property
    def dataset_size(self):
        return self.config.dataset_size

    @abstractmethod
    def _target_encoder(self):
        # Overridden by subclass to produce the right target encoding for a given target model.
        raise NotImplementedError

    def feed_shape_type_def(self):
        TS = tf.TensorShape
        if self.config.use_auxiliary_info:
            return (
                (
                    {"tokens": tf.int32, "mask": tf.float32, "context": tf.float32},
                    tf.float32,
                ),
                (
                    {
                        "tokens": TS([self.config.max_length, 2]),
                        "mask": TS([self.config.max_length]),
                        "context": TS([self.config.max_length, self.config.context_dim]),
                    },
                    TS([self.target_dim]),
                ),
            )
        else:
            return (
                ({"tokens": tf.int32, "mask": tf.float32}, tf.float32),
                (
                    {
                        "tokens": TS([self.config.max_length, 2]),
                        "mask": TS([self.config.max_length]),
                    },
                    TS([self.target_dim]),
                ),
            )

    def _array_format(self, encoded_output, pad_token=None):
        """
        Returns numpy array of token idxs and corresponding mask
        Returned `x` array contains two channels:
            0: byte-pair encoding embedding
            1: positional embedding
        """
        seq_length = len(encoded_output.token_ids)
        x = np.zeros((self.config.max_length, 2), dtype=np.int32)
        if self.config.base_model.__name__ == "RoBERTa":
            x += 1
        mask = np.zeros((self.config.max_length), dtype=np.float32)

        if encoded_output.labels is not None:
            labels_arr = np.empty((self.config.max_length), dtype="object")
            labels_arr.fill((pad_token or self.config.pad_token))
        else:
            labels_arr = None

        if encoded_output.context is not None:
            context_arr = np.zeros(
                (self.config.max_length, self.config.context_dim), dtype=np.float32
            )
        else:
            context_arr = None

        # BPE embedding
        x[:seq_length, 0] = encoded_output.token_ids
        # masking: value of 1 means "consider this in cross-entropy LM loss"
        mask[1:seq_length] = 1
        if encoded_output.labels:
            labels_arr[:seq_length] = encoded_output.labels
        if encoded_output.context is not None:
            if len(np.shape(encoded_output.context)) in (2, 3):
                context_arr[:seq_length][:] = np.squeeze(encoded_output.context)
            else:
                raise FinetuneError("Incorrect context rank.")

        # positional_embeddings
        x[:, 1] = np.arange(
            self.text_encoder.vocab_size,
            self.text_encoder.vocab_size + self.config.max_length,
        )

        # roberta uses different positional embedding structure
        if self.config.base_model.__name__ == "RoBERTa":
            mask = np.zeros((self.config.max_length), dtype=np.float32)
            mask[0:seq_length] = 1
            positions = np.cumsum(mask, dtype=np.int32)
            positions += 1  # add padding idx because RoBERTa's pos embeds depend on it
            positions += (
                self.text_encoder.vocab_size + 1
            )  # + 1 to include unused mask token in embedding layer
            x[:, 1] = positions

        output = ArrayEncodedOutput(
            token_ids=x,
            tokens=encoded_output.tokens,
            labels=labels_arr,
            char_locs=encoded_output.char_locs,
            mask=mask,
            context=context_arr,
        )
        return output

    def text_to_tokens_mask(self, X, Y=None, context=None):
        if context is None and self.config.use_auxiliary_info:
            context = X[1]
            X = X[0]

        out_gen = self._text_to_ids(X, pad_token=self.config.pad_token, context=context)

        for out in out_gen:
            if self.config.use_auxiliary_info:
                feats = {
                    "tokens": out.token_ids,
                    "mask": out.mask,
                    "context": out.context,
                }
            else:
                feats = {"tokens": out.token_ids, "mask": out.mask}
            if Y is None:
                yield feats
            else:
                yield feats, self.label_encoder.transform([Y])[0]

    def _post_data_initialization(self, Y=None, context=None):
        if Y is not None:
            self.label_encoder = self._target_encoder()
            if not callable(Y):
                Y_fit = Y
                self.label_encoder.fit(Y)
            else:
                Y_fit = list(itertools.islice(Y(), 10000))
                self.label_encoder.fit(Y_fit)
            self.config.pad_idx = self.pad_idx

            target_dim = self.label_encoder.target_dim
            self.lm_loss_coef = (
                self.config.lm_loss_coef if target_dim is not None else 1.0
            )
            self.target_dim = target_dim

        if context:
            if not callable(Y):
                characteristics = context
            else:
                characteristics = itertools.islice(context(), 10000)
            _ = self._context_to_vector(
                characteristics
            )  # to fit the auxiliary information label encoders

    def _context_to_vector(self, context):
        """
        Takes list of context dictionaries and turns them into lists (per sample) of lists (per token) of context vectors
        """
        num_samples = len(context)
        vector_list = []
        ignore = ["start", "end", "label", "token"]

        if not hasattr(
            self, "label_encoders"
        ):  # we will fit necessary label encoders here to know dimensionality of input vector before entering featurizer
            valid_samples = [
                dict_list for dict_list in context if dict_list != self.config.pad_token
            ]  # one list of dicts per sample of text
            pads_removed = [
                dictionary
                for dict_list in valid_samples
                for dictionary in dict_list
                if dictionary != self.config.pad_token
            ]  # concat each dictionary (per token) into a list of dict
            keys = [
                k
                for k in pads_removed[0].keys()
                if k not in ignore and k in self.config.default_context.keys()
            ]
            characteristics = {
                k: [dictionary[k] for dictionary in pads_removed] for k in keys
            }
            self.label_encoders = {
                k: LabelBinarizer()
                for k in characteristics.keys()
                if all(
                    k != self.config.pad_token
                    and not (
                        type(data) == int or (type(data) == float and not pd.isna(data))
                    )
                    and k not in ignore
                    for data in characteristics[k]
                )
            }  # Excludes features that are continuous, like position and font size, since they do not have categorical binary encodings

            for label, encoder in self.label_encoders.items():  # fit encoders
                without_nans = [str(x) for x in characteristics[label] if not pd.isna(x)]
                encoder.fit(without_nans)

            self.context_labels = sorted(
                [label for label in characteristics.keys() if label != "label"]
            )  # sort for consistent ordering between runs
            continuous_labels = [
                label
                for label in self.context_labels
                if label not in self.label_encoders.keys() and label not in ignore
            ]

            self.label_stats = {}
            for label in continuous_labels:
                stats = {
                    "mean": np.mean(characteristics[label]),
                    "std": np.std(characteristics[label]),
                }
                self.label_stats[label] = stats

            self.config.context_dim = len(
                continuous_labels
            )  # Sum over features to determine dimensionality of each feature vector.

            for (
                encoder
            ) in (
                self.label_encoders.values()
            ):  # since categorical labels use LabelBinarizer, they use varying dimensions each for each one-hot-encoding, while numerical features only use 1 dimension, so we sum over all for the total dimension.
                self.config.context_dim += (
                    1 if len(encoder.classes_) <= 2 else len(encoder.classes_)
                )  # Binary encodings with only two classes are given with one bit. Encodings with n > 2 classes are given with n bits. (Thanks sklearn)
            return True

        for sample in context:
            # See which tokens have padded labels/context vectors, and thus have a zero vector for features
            padded_indices = []
            tokens_with_context = []
            for idx in range(len(sample)):
                if type(sample[idx]) == str and sample[idx] == self.config.pad_token:
                    padded_indices.append(idx)
                else:
                    if not set(self.config.default_context.keys()).issubset(set(sample[idx].keys())):
                        missing = str(
                            set(self.config.default_context.keys()).difference(set(sample[idx].keys()))
                        )

                        raise FinetuneError(
                            "Token '{}' does not contain field(s) {} as described in default".format(
                                sample[idx]["token"], missing
                            )
                        )
                    tokens_with_context.append(sample[idx])
            characteristics = {
                k: [dictionary[k] for dictionary in tokens_with_context]
                for k in self.config.default_context.keys()
            }

            # make sure all features cover the same number of tokens, and calculate total num tokens
            num_tokens = None
            for label in self.context_labels:
                new_length = (
                    len(characteristics[label])
                    if type(characteristics[label] == list)
                    else 1
                )
                if num_tokens is not None and num_tokens != new_length:
                    raise FinetuneError("Incorrect label shapes.")
                num_tokens = new_length

            vector = np.zeros(
                (num_tokens + len(padded_indices), self.config.context_dim), dtype=np.float32
            )  # Feature vector for one document. Add 2 for the special tokens at beginning/end
            current_index = 0

            # Loop through each feature and add each to new index of the feature vector
            for label in self.context_labels:
                if label == self.config.pad_token or label in ignore:
                    continue
                data = characteristics[label]

                # Binary encoded features have different dimensionality as simple floats/ints, so must be handled differently.
                if label in self.label_encoders.keys():
                    without_nans = [x for x in data if not pd.isna(x)]
                    data = (
                        np.asarray(self.label_encoders[label].transform(without_nans))
                        - 0.5
                    )  # this removes nans from padded tokens, but now the list is too short. The 'num_backward' variable will track this offset to ensure indices are correctly tracked.
                    data_dim = len(self.label_encoders[label].classes_)
                    if data_dim == 2:  # since binary classes default to column vector
                        data_dim = 1
                else:  # need to normalize our float/int inputs
                    data = [str(x) for x in data if not pd.isna(x)]
                    stats = self.label_stats[label]
                    data = (data - stats["mean"]) / stats["std"]
                    data_dim = 1

                # loop through indices and fill with correct data
                tokens_added = 0
                for sample_idx in range(len(sample)):
                    if (
                        sample_idx in padded_indices
                    ):  # there is no data, simply a pad from the encoder, so fill out with zero vector
                        vector[sample_idx][:] = 0
                        continue
                    for label_dimension in range(data_dim):
                        vector[sample_idx][current_index + label_dimension] = (
                            data[tokens_added]
                            if data_dim == 1
                            else data[tokens_added][label_dimension]
                        )
                    tokens_added += 1
                current_index += 1
            vector_list.append(vector)
        return vector_list

    def _compute_class_counts(self, encoded_dataset):
        target_arrs = np.asarray([target_arr for doc, target_arr in encoded_dataset])
        return Counter(self.label_encoder.inverse_transform(target_arrs))

    def _dataset_with_targets(self, Xs, Y, train, context=None):
        if not callable(Xs) and not callable(Y):
            if self.config.use_auxiliary_info:
                dataset = lambda: zip(
                    Xs, Y, context
                )  # Do not need to check if context is callable - it is turned in along with Xs, and thus must have the same form
            else:
                dataset = lambda: zip(Xs, Y)
        elif callable(Xs) and callable(Y):
            if self.config.use_auxiliary_info:
                dataset = lambda: zip(
                    Xs(), Y(), context()
                )  # encode one sample at a time.
            else:
                dataset = lambda: zip(Xs(), Y())
        else:
            raise ValueError(
                "Either neither or both of Xs and Y should be callable, not a mixture"
            )

        dataset_encoded = lambda: itertools.chain.from_iterable(
            map(lambda xy: self.text_to_tokens_mask(*xy), dataset())
        )

        if not callable(Y) and train:
            dataset_encoded_list = list(dataset_encoded())
            class_counts = self._compute_class_counts(dataset_encoded_list)
            self.config.dataset_size = len(dataset_encoded_list)
            if self.config.class_weights is not None:
                self.config.class_weights = compute_class_weights(
                    class_weights=self.config.class_weights, class_counts=class_counts
                )
        shape_def = self.feed_shape_type_def()
        return Dataset.from_generator(
            lambda: self.wrap_tqdm(dataset_encoded(), train), *shape_def
        )

    def _dataset_without_targets(self, Xs, train, context=None):
        if not callable(Xs):
            if context:
                Xs = list(zip(Xs, context))
            Xs_fn = lambda: self.wrap_tqdm(Xs, train)
        else:
            Xs_fn = lambda: self.wrap_tqdm(Xs(), train)

        dataset_encoded = lambda: itertools.chain.from_iterable(
            map(self.text_to_tokens_mask, Xs_fn())
        )
        if not callable(Xs) and self.config.chunk_long_sequences:
            # Adjust dataset size to account for long documents being chunked
            dataset_encoded_list = list(dataset_encoded())
            self.config.dataset_size = len(dataset_encoded_list)
        types, shapes = self.feed_shape_type_def()
        return Dataset.from_generator(
            dataset_encoded, types[0], shapes[0]
        )  # 0s cut out the targets

    def _integer_val_size(self, val_size, dataset_size):
        if isinstance(val_size, float):
            return int(val_size * dataset_size)
        return val_size

    def validation_settings(self, n_examples, batch_size):
        """
        Auto-select reasonable validation settings
        """
        if self.config.val_size is not None and self.config.val_interval is not None:
            return (
                self._integer_val_size(self.config.val_size, n_examples),
                self.config.val_interval,
            )

        # Auto-select reasonable validation size
        if self.config.val_size == 'auto':
            if n_examples < 50 and not self.config.keep_best_model:
                val_size = 0
            else:
                val_size = max(5, int(0.05 * n_examples))
                val_size = min(100, val_size)
        else:
            val_size = self._integer_val_size(self.config.val_size, n_examples)

        # Auto-select reasonable validation interval
        if self.config.val_interval is None:
            # sys.maxsize corresponds to never running validation
            # and is used when val_size is set to 0
            val_interval = 4 * int(math.ceil(val_size / batch_size)) or sys.maxsize
        else:
            val_interval = self.config.val_interval

        return int(val_size), int(val_interval)

    def resampling(self, Xs, Y):
        return Xs, Y

    def _make_dataset(self, Xs, Y, context=None, train=False):
        if Y is not None:
            dataset = lambda: self._dataset_with_targets(
                Xs, Y, context=context, train=train
            )
        else:
            dataset = lambda: self._dataset_without_targets(
                Xs, context=context, train=train
            )
        return dataset

    def wrap_tqdm(self, gen, train):
        if self.config.debugging_logs:
            return gen

        if train is None:
            return gen

        try:
            total = len(gen)
        except:
            if train:
                total = self.config.dataset_size
            else:
                total = self.config.val_size

        def internal_gen():
            current_epoch = (self.epoch - 1) % self.config.n_epochs + 1
            it = iter(gen)

            if train:
                if self.config.prefit_init and self.epoch <= self.config.n_epochs:
                    desc = "Initialization Epoch {}/{}".format(
                        current_epoch, self.config.n_epochs
                    )
                else:
                    desc = "Epoch {}/{}".format(current_epoch, self.config.n_epochs)
            else:
                desc = "Validation"
            for _, i in zip(range(self._skip_tqdm), it):
                yield i

            for i in tqdm.tqdm(
                it,
                desc=desc,
                total=total,
                miniters=1,
                leave=current_epoch == self.config.n_epochs and train,
            ):
                yield i

            if train:
                self.epoch += 1

        return internal_gen()

    def get_train_input_fns(
        self, Xs, Y=None, context=None, batch_size=None, val_size=None
    ):
        self.epoch = 1
        batch_size = batch_size or self.config.batch_size

        shuffle_buffer_size = self.config.shuffle_buffer_size
        val_size = val_size or 0
        prefetch_buffer = 2  # breaks the pipeline to allow concurrency

        if callable(Xs):
            try:
                self.config.dataset_size = len(Xs())
            except TypeError:
                if self.config.dataset_size is None:
                    raise FinetuneError(
                        "Generator input function does not have a length and no `config.dataset_size` is specified. "
                        "You must set `config.dataset_size` explicitly."
                    )
        else:
            self.config.dataset_size = len(Xs)

        self.config.val_size, self.config.val_interval = self.validation_settings(
            n_examples=len(Xs) if not callable(Xs) else self.config.dataset_size,
            batch_size=batch_size or self.config.batch_size,
        )
        self.config.dataset_size -= val_size

        context_style = None
        if context:
            context_style = context
        if Y is not None:
            self._post_data_initialization(Y=Y, context=context_style)
        else:
            self._post_data_initialization(Y=None, context=context_style)

        if callable(Xs) or Y is None:
            self._skip_tqdm = val_size
            dataset = self._make_dataset(Xs, Y, context, train=True)
            val_dataset_unbatched = (
                lambda: dataset()
                .shuffle(
                    shuffle_buffer_size,
                    seed=self.config.seed,
                    reshuffle_each_iteration=False,
                )
                .take(self.config.val_size)
            )
            train_dataset_unbatched = (
                lambda: dataset()
                .shuffle(
                    shuffle_buffer_size,
                    seed=self.config.seed,
                    reshuffle_each_iteration=False,
                )
                .skip(self.config.val_size)
            )
        else:
            self._skip_tqdm = 0
            if self.config.val_set is None:
                if self.config.val_size == 0:
                    Xs_tr, Xs_va, Y_tr, Y_va, C_tr, C_va = Xs, [], Y, [], context, []
                else:
                    if context:
                        raise FinetuneError(
                            "Validation set with auxiliary info not yet supported."
                        )
                        Xs_tr, Xs_va, Y_tr, Y_va, C_tr, C_va = train_test_split(
                            Xs,
                            Y,
                            context,
                            test_size=self.config.val_size,
                            random_state=self.config.seed,
                        )
                    else:
                        Xs_tr, Xs_va, Y_tr, Y_va = train_test_split(
                            Xs,
                            Y,
                            test_size=self.config.val_size,
                            random_state=self.config.seed,
                        )
                        C_tr, C_va = None, None
            else:
                Xs_tr, Y_tr, C_tr = Xs, Y, context
                if context:
                    Xs_va, Y_va, C_va = self.config.val_set
                else:
                    Xs_va, Y_va, C_va = self.config.val_set

            Xs_tr, Y_tr = self.resampling(Xs_tr, Y_tr)
            self.config.dataset_size = len(Xs_tr)
            val_dataset_unbatched = self._make_dataset(Xs_va, Y_va, C_va, train=False)
            train_dataset_unbatched = self._make_dataset(Xs_tr, Y_tr, C_tr, train=True)

        if self.config.chunk_long_sequences or self.config.class_weights:
            # Certain settings require that the entire dataset be encoded before compiling the graph
            train_dataset_unbatched()

        val_dataset = (
            lambda: val_dataset_unbatched()
            .batch(batch_size, drop_remainder=False)
            .cache()
            .prefetch(prefetch_buffer)
        )
        train_dataset = (
            lambda: train_dataset_unbatched()
            .batch(batch_size, drop_remainder=False)
            .repeat(self.config.n_epochs)
            .prefetch(prefetch_buffer)
        )

        return (
            val_dataset,
            train_dataset,
            self.config.val_size,
            self.config.val_interval,
        )

    def get_predict_input_fn(self, Xs, batch_size=None, context=None):
        batch_size = batch_size or self.config.batch_size
        tf_dataset = lambda: self._dataset_without_targets(
            Xs, train=None, context=context
        ).batch(batch_size)
        return tf_dataset

    @property
    def pad_idx(self):
        if self.pad_idx_ is None:
            if hasattr(self.label_encoder, "classes_"):
                classes = list(self.label_encoder.classes_)
                if self.config.pad_token in classes:
                    self.pad_idx_ = classes.index(self.config.pad_token)
                else:
                    self.pad_idx_ = None
        return self.pad_idx_

    def _format_for_encoding(self, X):
        """
        Most subclasses take in inputs as:
            List (batch) of list (docs)

        Encode_multi_input expect the following format:
            List (batch) of list (docs) of list (subseqs) of text

        This method is responsible for standardizing inputs to the above format
        """
        return [[X]]

    def _format_for_inference(self, X):
        return list(X)

    def _text_to_ids(self, Xs, Y=None, pad_token=None, context=None):
        if context is None and self.config.use_auxiliary_info:
            context = Xs[0]
            Xs = Xs[1]
        Xs = self._format_for_encoding(Xs)
        if self.config.chunk_long_sequences and len(Xs) == 1:
            # can only chunk single sequence inputs
            chunk_size = self.config.max_length - 2
            step_size = chunk_size // 3
            encoded = self.text_encoder.encode_multi_input(
                Xs,
                Y=Y,
                max_length=sys.maxsize,
                pad_token=(pad_token or self.config.pad_token),
                context=context,
            )
            if self.config.use_auxiliary_info:
                processed_context = np.squeeze(
                    self._context_to_vector([encoded.context])
                )

            length = len(encoded.token_ids)
            assert length == len(encoded.token_ids)
            starts = list(range(0, length, step_size))
            for start in starts:
                d = dict()
                end = start + chunk_size
                for field in EncodedOutput._fields:
                    field_value = getattr(encoded, field)
                    if field_value is not None:
                        d[field] = field_value[start:end]
                if self.config.use_auxiliary_info:
                    d["context"] = processed_context[
                        start:end
                    ]  # forced since encoded is immutable'
                else:
                    d['context'] = None

                yield self._array_format(EncodedOutput(**d), pad_token=pad_token)
        else:
            encoder_out = self.text_encoder.encode_multi_input(
                Xs,
                Y=Y,
                max_length=self.config.max_length,
                pad_token=(pad_token or self.config.pad_token),
                context=context,
            )

            d = dict()
            for field in EncodedOutput._fields:
                field_value = getattr(encoder_out, field)
                if field_value is not None:
                    d[field] = field_value
            if self.config.use_auxiliary_info:
                d["context"] = np.squeeze(
                    self._context_to_vector([encoder_out.context])
                )  # forced since encoded is immutable
            else:
                d["context"] = None

            yield self._array_format(
                EncodedOutput(**d), pad_token=(pad_token or self.config.pad_token)
            )

